@@ -17,7 +17,7 @@
 
 #include "Rmpi.h"
 
-#ifndef Win32
+#ifdef OPENMPI
 #include <dlfcn.h>
 #endif
 
@@ -27,13 +27,41 @@ static MPI_Datatype *datatype;
 static MPI_Info *info;
 static MPI_Request *request;
 static int COMM_MAXSIZE=10;
-static int STATUS_MAXSIZE=5000;
-static int REQUEST_MAXSIZE=10000;
+static int STATUS_MAXSIZE=2000;
+static int REQUEST_MAXSIZE=2000;
+static MPI_Datatype *xdouble;
 
-SEXP mpi_initialize(){
+#ifndef XLENGTH
+#define XLENGTH LENGTH
+#endif
+
+SEXP mpidist(void){
+	int i=0;
+
+#ifdef OPENMPI
+	i=1;
+#endif
+
+#ifdef LAM
+        i=2;
+#endif
+
+#ifdef MPICH
+	i=3;
+#endif
+
+#if defined(MPICH2) || defined(INTELMPI)
+	i=4;
+#endif
+
+	return AsInt(i);	
+}
+
+SEXP mpi_initialize(void){
 	int i,flag;
 	MPI_Initialized(&flag);
 
+
 #ifndef MPI2
         static int fake_argc = 1;
        	char *fake_argv[1];
@@ -43,23 +71,31 @@ SEXP mpi_initialize(){
 if (flag)
 	return AsInt(1);
 	else {
-	
+
+#ifndef __APPLE__
 #ifdef OPENMPI
-	dlopen("libmpi.so.0", RTLD_GLOBAL);
+    if (!dlopen("libmpi.so.1", RTLD_GLOBAL | RTLD_LAZY) 
+	&& !dlopen("libmpi.so.0", RTLD_GLOBAL | RTLD_LAZY)
+	&& !dlopen("libmpi.so", RTLD_GLOBAL | RTLD_LAZY)) {
+        Rprintf("%s\n",dlerror());
+        return AsInt(0);
+    }
+#endif
 #endif
 
 #ifndef MPI2
    	fake_argv[0] = (char *)&fake_argv0;
        	MPI_Init(&fake_argc, (char ***)(void*)&fake_argv);
-#else
+#else 
 	MPI_Init((void *)0,(void *)0);
-#endif
+#endif 
 
-		MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
-		MPI_Errhandler_set(MPI_COMM_SELF, MPI_ERRORS_RETURN);
+		MPI_Comm_set_errhandler(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
+		MPI_Comm_set_errhandler(MPI_COMM_SELF, MPI_ERRORS_RETURN);
 		comm=(MPI_Comm *)Calloc(COMM_MAXSIZE, MPI_Comm); 
 		status=(MPI_Status *)Calloc(STATUS_MAXSIZE, MPI_Status); 
 		datatype=(MPI_Datatype *)Calloc(1, MPI_Datatype); 
+		xdouble=(MPI_Datatype *)Calloc(1, MPI_Datatype); 
 		info=(MPI_Info *)Calloc(1, MPI_Info);
 		info[0]=MPI_INFO_NULL;
 		request=(MPI_Request *)Calloc(REQUEST_MAXSIZE, MPI_Request);
@@ -68,20 +104,21 @@ if (flag)
 		for (i=1;i < COMM_MAXSIZE; comm[i++]=MPI_COMM_NULL);
 
 		return AsInt(1);
-	}
+	} 
 }
 
-SEXP mpi_finalize(){
+SEXP mpi_finalize(void){
 	MPI_Finalize();
 	Free(comm);
 	Free(status);
 	Free(request);
 	Free(datatype);
+	Free(xdouble);
 	Free(info);
 	return AsInt(1);
 }
 
-SEXP mpi_get_processor_name (){
+SEXP mpi_get_processor_name (void){
 	int resultlen;
 	char *name;
 	SEXP sexp_name;
@@ -95,12 +132,14 @@ SEXP mpi_get_processor_name (){
 	return sexp_name;
 }
 
+/*
 SEXP bin_nchar(SEXP sexp_data){
 	return AsInt(LENGTH(STRING_ELT(sexp_data,0)));
 }
+*/
 
 #ifdef MPI2
-SEXP mpi_universe_size(){
+SEXP mpi_universe_size(void){
 	int *MPI_Universe_Size;
 	int univ_flag;
 	MPI_Comm_get_attr(comm[0], MPI_UNIVERSE_SIZE, &MPI_Universe_Size, &univ_flag);
@@ -111,19 +150,19 @@ SEXP mpi_universe_size(){
 }
 #endif
 
-SEXP mpi_any_source(){
+SEXP mpi_any_source(void){
 	return AsInt(MPI_ANY_SOURCE);
 }
 
-SEXP mpi_any_tag(){
+SEXP mpi_any_tag(void){
 	return AsInt(MPI_ANY_TAG);
 }
 
-SEXP mpi_undefined(){
+SEXP mpi_undefined(void){
 	return AsInt(MPI_UNDEFINED);
 }
 
-SEXP mpi_proc_null(){
+SEXP mpi_proc_null(void){
 	return AsInt(MPI_PROC_NULL);
 }
 
@@ -166,7 +205,7 @@ SEXP mpi_realloc_comm(SEXP sexp_newncomm){
 	return AsInt(1);
 }
 
-SEXP mpi_comm_maxsize(){
+SEXP mpi_comm_maxsize(void){
 	return AsInt(COMM_MAXSIZE);
 }
 
@@ -179,7 +218,7 @@ SEXP mpi_realloc_status(SEXP sexp_newnstatus){
 	return AsInt(1);
 }
 
-SEXP mpi_status_maxsize(){
+SEXP mpi_status_maxsize(void){
 	return AsInt(STATUS_MAXSIZE);
 }
 
@@ -193,7 +232,7 @@ SEXP mpi_realloc_request(SEXP sexp_newnrequest){
 	return AsInt(1);
 }
 
-SEXP mpi_request_maxsize(){
+SEXP mpi_request_maxsize(void){
 	return AsInt(REQUEST_MAXSIZE);
 }
 
@@ -554,14 +593,17 @@ SEXP mpi_allgatherv(SEXP sexp_sdata,
 SEXP mpi_bcast(SEXP sexp_data,
 			   SEXP sexp_type,
 			   SEXP	sexp_rank,
-			   SEXP sexp_comm){
+			   SEXP sexp_comm,
+			   SEXP sexp_buffunit){
 
 	int len=LENGTH(sexp_data), type=INTEGER(sexp_type)[0];
 	int rank=INTEGER(sexp_rank)[0], root,  commn=INTEGER(sexp_comm)[0],slen;
-	int errcode=0;
+	int buffunit=INTEGER(sexp_buffunit)[0],errcode=0;
 	char *rdata;
 	SEXP sexp_data2 = NULL;
-
+	//MPI_Datatype xdouble;
+	R_xlen_t xlen=XLENGTH(sexp_data);
+	
 	switch (type){
 	case 1:
 		errcode=MPI_Bcast(INTEGER(sexp_data), len, MPI_INT, rank, comm[commn]);
@@ -587,7 +629,13 @@ SEXP mpi_bcast(SEXP sexp_data,
 	case 4:
                 errcode=MPI_Bcast(RAW(sexp_data), len, MPI_BYTE, rank, comm[commn]);
                 break;
-
+	case 5:
+		MPI_Type_contiguous(buffunit, MPI_DOUBLE, xdouble);
+		MPI_Type_commit(xdouble);
+		if ((xlen % buffunit) > 0) len=1+(xlen/buffunit); else len=xlen/buffunit;
+        mpi_errhandler(MPI_Bcast(REAL(sexp_data), len, xdouble[0], rank, comm[commn]));
+		MPI_Type_free(xdouble);
+		break;
 	default:
 		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
 		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
@@ -936,7 +984,7 @@ SEXP mpi_comm_rank(SEXP sexp_comm){
 
 SEXP mpi_comm_dup(SEXP sexp_comm, SEXP sexp_newcomm){
     int commn=INTEGER(sexp_comm)[0], newcommn=INTEGER(sexp_newcomm)[0];
-    if (comm==0)
+    if (commn==0)
         return AsInt(erreturn(mpi_errhandler(MPI_Comm_dup(MPI_COMM_WORLD,
                 &comm[newcommn]))));
     else
@@ -944,6 +992,11 @@ SEXP mpi_comm_dup(SEXP sexp_comm, SEXP sexp_newcomm){
                 &comm[newcommn]))));
 }
 
+SEXP mpi_comm_c2f(SEXP sexp_comm){
+  int c = INTEGER(sexp_comm)[0];
+  return AsInt(MPI_Comm_c2f(comm[c]));
+}
+
 SEXP mpi_comm_free(SEXP sexp_comm){
 	return AsInt(erreturn(mpi_errhandler(MPI_Comm_free(&comm[INTEGER(sexp_comm)[0]]))));
 }
@@ -960,7 +1013,7 @@ SEXP mpi_abort(SEXP sexp_comm){
 
 /********************Intercomm********************************************/
 SEXP mpi_comm_set_errhandler(SEXP sexp_comm){
-	return AsInt(erreturn(MPI_Errhandler_set(comm[INTEGER(sexp_comm)[0]], 
+	return AsInt(erreturn(MPI_Comm_set_errhandler(comm[INTEGER(sexp_comm)[0]], 
 		MPI_ERRORS_RETURN)));
 }
 
@@ -976,10 +1029,12 @@ SEXP mpi_comm_spawn (SEXP sexp_slave,
 					 SEXP sexp_nslave,
 					 SEXP sexp_info,
 					 SEXP sexp_root,
-					 SEXP sexp_intercomm){
+					 SEXP sexp_intercomm,
+					 SEXP sexp_quiet){
     int i, nslave = INTEGER (sexp_nslave)[0], len = LENGTH (sexp_argv);
 	int infon=INTEGER(sexp_info)[0], root=INTEGER(sexp_root)[0];
 	int intercommn=INTEGER(sexp_intercomm)[0], *slaverrcode, realns;
+    int quiet = INTEGER(sexp_quiet)[0];
 
 	slaverrcode = (int *)Calloc(nslave, int);
 	if (len==0)
@@ -1001,8 +1056,8 @@ SEXP mpi_comm_spawn (SEXP sexp_slave,
 		for (i=0; i < nslave; mpi_errhandler(slaverrcode[i++]));
 
 	Free(slaverrcode);
-	
-	Rprintf("\t%d slaves are spawned successfully. %d failed.\n", realns, nslave-realns);
+	if (!quiet || realns < nslave)
+		Rprintf("\t%d slaves are spawned successfully. %d failed.\n", realns, nslave-realns);
     return AsInt(realns);
 }
 
@@ -1010,7 +1065,7 @@ SEXP mpi_comm_get_parent(SEXP sexp_comm){
 	return AsInt(erreturn(mpi_errhandler(MPI_Comm_get_parent(&comm[INTEGER(sexp_comm)[0]]))));
 }
 
-SEXP mpi_is_master(){
+SEXP mpi_is_master(void){
 	int check;
 	MPI_Comm master;
 	MPI_Comm_get_parent(&master);
@@ -1487,3 +1542,4 @@ SEXP mpi_request_get_status(SEXP sexp_request,  SEXP sexp_status){
         return AsInt(flag);
 }
 */
+
